//********************************************************************
//*** ResourcePool/ResourcePoolOptions.hh
//*** Copyright (c) 2003-2004 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePoolOptions.hh,v 1.17 2004/11/17 14:39:12 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_RESOURCEPOOLOPTIONS_HH
#define RESOURCEPOOL_RESOURCEPOOLOPTIONS_HH

#ifndef POSIX_TIME_HH
#include "POSIX/time.hh"
#endif

#ifndef INCLUDED_EXCEPTIONS_HH
#include "Exceptions/Exception.hh"
#endif

#ifndef INCLUDED_THREAD_LOCK_HH
#include "Thread/Lock.hh"
#endif

#include <vector>
#include <string>

namespace fatalmind {
/**
 * Options configuration for ResourcePool.
 */
class ResourcePoolOptions {
		typedef POSIX::timespec<SingleThreadedModel<> > timespec_t;
		typedef std::vector<POSIX::timespec<> >         SleepOnFail_t;

		SleepOnFail_t _SleepOnFail;
		int           _Max;
		int           _MaxTry;
		int           _MaxExecTry;
		int           _PreCreate;
		timespec_t    _wait;	

	public:
		ResourcePoolOptions();	
		~ResourcePoolOptions() {
		}

		/**
		 * Specifies the maximum concurrent resources managed by this
		 * pool. If the limit is reached the ResourcePool::get()
		 * method will throw an Exception
		 */
		void setMax(const int a_Max);

		/**
		 * Specifies how many dead resources the ResourcePool::get()
		 * method checks before it throws an Exception. 
		 * Normally the ResourcePool::get() method doesn't
		 * return dead resources (e.g. broken connections). In the
		 * case there is a broken connection in the pool, the ResourcePool::get()
		 * method throws it away and takes the next resource out of
		 * the pool. The ResourcePool::get() method tries not more then MaxTry
		 * resources before it throws an Exception.
		 */
		void setMaxTry(const int a_MaxTry);

		/**
		 * Specifies how often the ResourcePool::execute() method will 
		 * repeate the command before it gives up. Please have a look at the
		 * Command documentation to lern more about the
		 * Command patten and how to use it with ResourcePool.
		 */
		void setMaxExecTry(const int a_MaxExecTry);

		/**
		 * How many resource to create on startup.
		 * Normally the ResourcePool creates new resources only on
		 * demand, with this option you can specify how many
		 * resources are created in advance when the ResourcePool
		 * gets constructed.
		 */
		void setPreCreate(const int a_PreCreate);

		/**
		 * Delay if dead resources are detected.
		 * You can tell the ResourcePool to sleep a while when a dead
		 * resource was found. Normally the ResourcePool::get() 
		 * method tries up to MaxTry times to get a valid resource for you 
		 * without any delay between the attempts. This is usually not what 
		 * you want if you database is not available for a short time. 
		 * Using this option you can specify a list of timeouts which should be
		 * slept between two attempts to get an valid resource.
		 *
		 * If you have specified
		 * - MaxTry = 5
		 * - SleepOnFail = [0, 1, 2, 4]
		 *
		 * ResourcePool would do the following if it isn't able to
		 * get an valid resource:
		 * - try to get a resource: fails
		 * - sleep 0 seconds
		 * - try to get a resource: fails
		 * - sleep 1 second
		 * - try to get a resource: fails
		 * - sleep 2 seconds
		 * - try to get a resource: fails
		 * - sleep 4 seconds
		 * - try to get a resource: fails
		 * - give up and throw an Exception
		 * 
 		 * Using an exponential time scheme like this one, is usually
		 * the most efficient. However it is highly recommended to
		 * leave the first value always "0" since this is required to
		 * allow the ResourcePool to try to establish a new connection 
		 * without delay.
		 *
		 * The number of sleeps can not be more than MaxTry - 1, if you
		 * specify more values the list is truncated. If you specify
		 * less values the list is extended using the last value for
		 * the extended elements. e.g. [0, 1] in the above example
		 * would have been extended to [0, 1, 1, 1]
		 *
		 */
		void setSleepOnFail(const SleepOnFail_t&);
		void setSleepOnFail(const float
			, const float = -1
			, const float = -1
			, const float = -1
		);
		void setWait();
		void setWait(const float);

		inline int getPreCreate() const {
			return _PreCreate;
		}

		inline int getMaxTry() const {
			return _MaxTry;
		}

		inline int getMaxExecTry() const {
			return _MaxExecTry;
		}

		inline int getMax() const {
			return _Max;
		}

		bool SleepOnFail(const int) const;

		POSIX::timespec<> getSleepOnFail(int tryno) const;

		const timespec_t* getWait() const {
			return &_wait;
		}

		class IllegalArgumentException: public fatalmind::Exception {
			public:
				IllegalArgumentException(const std::string& msg)
				: fatalmind::Exception(msg)
				{
				};
		};

		static const ResourcePoolOptions defaultOptions;
};
		
} // namespace fatalmind

#endif
