//********************************************************************
//*** ResourcePool/SQL/Types.t.cc
//*** Copyright (c) 2003-2004 by Markus Winand <mws@fatalmind.com>
//*** $Id: Types.t.cc,v 1.10 2004/11/17 14:39:15 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */


#include "t/main.hh"

using namespace std;
using namespace fatalmind;

#include <memory>
#include <iostream>
#include <limits>

std::auto_ptr<RP> _rp;

void test_init()
{
	if (_rp.get() == 0) {
		_rp.reset(new RP(getFactory()));
	}
}

void test_Types_init()
{
	test_init();
}

class TblCreate {
	public:
		TblCreate(RP& p, const std::string& create, const std::string& tbl) 
		: _p(p)
		, _tbl(tbl)
		{
			drop();
			RP::Command::SQLExecute cmd(create);
			_p.execute(cmd);
		}

		void drop() {
			try {
				std::string drop = "drop table ";
				drop.append(_tbl);
				RP::Command::SQLExecute cmd(drop);
				_p.execute(cmd);
			} catch (Exception e) {
				// TODO: oracle never fails here...hmmm...something strange with DDL???
				//std::cerr << "drop failed " << e.message() << std::endl;
			}
		}

		~TblCreate() {
			drop();
		}
	private:
		RP& _p;
		std::string _tbl;
};

template<typename T>
void test_int(int& id)
{
	int mxid = id++;
	int miid = id++;

	RP::Command::SQLExecute insert("insert into hirsch (id, val) values (?, ?)");
	T max_val = std::numeric_limits<T>::max();
	insert.bind(0, mxid);
	insert.bind(1, max_val);
	_rp->execute(insert);
	
	T min_val = std::numeric_limits<T>::min();
	insert.bind(0, miid);
	insert.bind(1, min_val);
	_rp->execute(insert);


	RP::Command::SQLSelectRow select("select val from hirsch where id = ?");
	T sel_val;
	select.bindout(0, sel_val);

	select.bind(0, mxid);
	_rp->execute(select);
	assertEqualT(T, sel_val, std::numeric_limits<T>::max());

	select.bind(0, miid);
	_rp->execute(select);
	assertEqualT(T, sel_val, std::numeric_limits<T>::min());

	RP::Command::SQLSelectRow select2("select val - 1 from hirsch where id = ?");
	select2.bind(0, mxid);
	select2.bindout(0, sel_val);
	_rp->execute(select2);
	assertEqualT(T, sel_val, (std::numeric_limits<T>::max()-1));

	try {
		select2.bind(0, miid);
		_rp->execute(select2);
		std::cerr << "FAIL: signed: " << std::numeric_limits<T>::is_signed << " size: " << sizeof(T) 
			<< " id max " << mxid << "  id min " << miid << std::endl;
		assertNotReach();
	} catch (SQL::SQLException e) {
	}
}

void test_Types_int()
{
	TblCreate tbl(*_rp
		, "create table hirsch (id numeric not null primary key, val numeric)"
		, "hirsch"
	);
	int id = 0;
	test_int<short int>(id);
	test_int<int>(id);
	test_int<long int>(id);

	test_int<unsigned short int>(id);
	test_int<unsigned int>(id);
	test_int<unsigned long int>(id);
}

template<typename T>
void test_float(int& id) 
{
	int nulid = id++;
	int tetid = id++;
	int temtid = id++;
	int mtemtid = id++;
	T val;

	RP::Command::SQLExecute insert("insert into hirsch (id, val) values (?, ?)");
	val = 0.0;
	insert.bind(0, nulid);
	insert.bind(1, val);
	_rp->execute(insert);
	
	insert.bind(0, tetid);
	val = 10E10;
	insert.bind(1, val);
	_rp->execute(insert);

	insert.bind(0, temtid);
	val = 10E-10;
	insert.bind(1, val);
	_rp->execute(insert);

	insert.bind(0, mtemtid);
	val = -10E-10;
	insert.bind(1, val);
	_rp->execute(insert);


	RP::Command::SQLSelectRow select("select val from hirsch where id = ?");
	T sel_val;
	select.bindout(0, sel_val);
	select.bind(0, nulid);
	_rp->execute(select);
	assertEqualT(T, sel_val, 0.0);

	select.bind(0, tetid);
	_rp->execute(select);
	assertEqualT(T, sel_val, 100000000000.0);

	select.bind(0, temtid);
	_rp->execute(select);
	assertEqualT(T, sel_val, 0.000000001);

	select.bind(0, mtemtid);
	_rp->execute(select);
	assertEqualT(T, sel_val, -0.000000001);

	RP::Command::SQLSelectRow select2("select val+? from hirsch where id = ?");
	select2.bind(0, std::numeric_limits<T>::epsilon());	
	select2.bind(1, nulid);
	select2.bindout(0, sel_val);
	_rp->execute(select2);
	assertEqualT(T, sel_val, std::numeric_limits<T>::epsilon());

	select2.bind(1, tetid);
	_rp->execute(select2);
	assertEqualT(T, sel_val, 100000000000.0 + std::numeric_limits<T>::epsilon());

	select2.bind(1, temtid);
	_rp->execute(select2);
	assertEqualT(T, sel_val, 0.000000001 + std::numeric_limits<T>::epsilon());

	select2.bind(1, mtemtid);
	_rp->execute(select2);
	assertEqualT(T, sel_val, -0.000000001 + std::numeric_limits<T>::epsilon());

}

void test_Types_float()
{
	int id = 12;
	{
		TblCreate tbl(*_rp
			, "create table hirsch (id numeric not null primary key, val float)"
			, "hirsch"
		);
		test_float<float>(id);
	}
}
void test_Types_double()
{
	// for some very strange reason, droping the talbe does not work, so i take a new session.
	// then it works. TODO: checkthis
	_rp.reset(0); test_init();
	int id = 14;
	{
		TblCreate tbl2(*_rp
			, "create table hirsch (id numeric not null primary key, val  DOUBLE PRECISION)"
			, "hirsch"
		);
		test_float<double>(id);
	}
}

void test_Types_string() 
{
	_rp.reset(0); test_init();
	TblCreate tbl2(*_rp
		, getStringTable()
		, "hirsch"
	);
	std::string val;
	for (int i = 0; i < 255; ++i) {
		val.append(1, char(i%256));
	}
	RP::Command::SQLExecute insert("insert into hirsch (id, val) values (?, ?)");
	insert.bind(0, 1);
	insert.bind(1, val);
	_rp->execute(insert);

	std::string out;
	RP::Command::SQLSelectRow get("select val from hirsch where id = ?");
	get.bind(0, 1);
	get.bindout(0, out);
	_rp->execute(get);

	assertEqual(val, out);
}

void priv_test_lob_row(size_t lenght, int id = 1, int offset = 0) {
	std::string val;
	for (size_t i = 0; i < lenght; ++i) {
		val.push_back(i%256);
	}
	RP::Command::SQLExecute insert("insert into hirsch (id, val) values (?, ?)");
	insert.bind(0, id);
	insert.bind(1, val);
	_rp->execute(insert);

	std::string out;
	RP::Command::SQLSelectRow get("select val from hirsch where id = ?");
	get.bind(0, id);
	get.bindout(0, out);
	_rp->execute(get);

	assertEqual(out, val);
}
void test_Types_lob() 
{
	_rp.reset(0); test_init();
	TblCreate tbl2(*_rp
		, getLobTable()
		, "hirsch"
	);
	priv_test_lob_row(4096, 1);
	priv_test_lob_row(65535, 2);
	priv_test_lob_row(1048568, 3);
}

void priv_test_perf_Types_lob_4k(int length) 
{
	_rp.reset(0); test_init();
	TblCreate tbl2(*_rp
		, getLobTable()
		, "hirsch"
	);
	for (int i = 0; i < 5; ++i) {
		priv_test_lob_row(length, i);
	}
}

void test_perf_Types_lob_4k() {
	priv_test_perf_Types_lob_4k(4096);
}

void test_perf_Types_lob_64k() {
	priv_test_perf_Types_lob_4k(65536);
}

void test_perf_Types_lob_1024k() {
	priv_test_perf_Types_lob_4k(1048568);
}

void test_Types_cleanup() 
{
//    drop(*_rp);
	_rp.reset(0);
}
